#include <cstdint>
#include <list>
#include <string>
#include <vector>

#include "common/common/fmt.h"
#include "common/stats/isolated_store_impl.h"

#include "extensions/filters/network/redis_proxy/command_splitter_impl.h"
#include "extensions/filters/network/redis_proxy/supported_commands.h"

#include "test/extensions/filters/network/redis_proxy/mocks.h"
#include "test/mocks/common.h"
#include "test/test_common/printers.h"

#include "gmock/gmock.h"
#include "gtest/gtest.h"

using testing::_;
using testing::ByRef;
using testing::DoAll;
using testing::Eq;
using testing::InSequence;
using testing::Ref;
using testing::Return;
using testing::WithArg;

namespace Envoy {
namespace Extensions {
namespace NetworkFilters {
namespace RedisProxy {
namespace CommandSplitter {

class RedisCommandSplitterImplTest : public testing::Test {
public:
  void makeBulkStringArray(RespValue& value, const std::vector<std::string>& strings) {
    std::vector<RespValue> values(strings.size());
    for (uint64_t i = 0; i < strings.size(); i++) {
      values[i].type(RespType::BulkString);
      values[i].asString() = strings[i];
    }

    value.type(RespType::Array);
    value.asArray().swap(values);
  }

  ConnPool::MockInstance* conn_pool_{new ConnPool::MockInstance()};
  Stats::IsolatedStoreImpl store_;
  InstanceImpl splitter_{ConnPool::InstancePtr{conn_pool_}, store_, "redis.foo."};
  MockSplitCallbacks callbacks_;
  SplitRequestPtr handle_;
};

TEST_F(RedisCommandSplitterImplTest, InvalidRequestNotArray) {
  RespValue response;
  response.type(RespType::Error);
  response.asString() = "invalid request";
  EXPECT_CALL(callbacks_, onResponse_(PointeesEq(&response)));
  RespValue request;
  EXPECT_EQ(nullptr, splitter_.makeRequest(request, callbacks_));

  EXPECT_EQ(1UL, store_.counter("redis.foo.splitter.invalid_request").value());
}

TEST_F(RedisCommandSplitterImplTest, InvalidRequestArrayTooSmall) {
  RespValue response;
  response.type(RespType::Error);
  response.asString() = "invalid request";
  EXPECT_CALL(callbacks_, onResponse_(PointeesEq(&response)));
  RespValue request;
  makeBulkStringArray(request, {"incr"});
  EXPECT_EQ(nullptr, splitter_.makeRequest(request, callbacks_));

  EXPECT_EQ(1UL, store_.counter("redis.foo.splitter.invalid_request").value());
}

TEST_F(RedisCommandSplitterImplTest, InvalidRequestArrayNotStrings) {
  RespValue response;
  response.type(RespType::Error);
  response.asString() = "invalid request";
  EXPECT_CALL(callbacks_, onResponse_(PointeesEq(&response)));
  RespValue request;
  makeBulkStringArray(request, {"incr", ""});
  request.asArray()[1].type(RespType::Null);
  EXPECT_EQ(nullptr, splitter_.makeRequest(request, callbacks_));

  EXPECT_EQ(1UL, store_.counter("redis.foo.splitter.invalid_request").value());
}

TEST_F(RedisCommandSplitterImplTest, UnsupportedCommand) {
  RespValue response;
  response.type(RespType::Error);
  response.asString() = "unsupported command 'newcommand'";
  EXPECT_CALL(callbacks_, onResponse_(PointeesEq(&response)));
  RespValue request;
  makeBulkStringArray(request, {"newcommand", "hello"});
  EXPECT_EQ(nullptr, splitter_.makeRequest(request, callbacks_));

  EXPECT_EQ(1UL, store_.counter("redis.foo.splitter.unsupported_command").value());
}

class RedisSingleServerRequestTest : public RedisCommandSplitterImplTest,
                                     public testing::WithParamInterface<std::string> {
public:
  void makeRequest(const std::string& hash_key, const RespValue& request) {
    EXPECT_CALL(*conn_pool_, makeRequest(hash_key, Ref(request), _))
        .WillOnce(DoAll(WithArg<2>(SaveArgAddress(&pool_callbacks_)), Return(&pool_request_)));
    handle_ = splitter_.makeRequest(request, callbacks_);
  }

  void fail() {
    RespValue response;
    response.type(RespType::Error);
    response.asString() = "upstream failure";
    EXPECT_CALL(callbacks_, onResponse_(PointeesEq(&response)));
    pool_callbacks_->onFailure();
  }

  void respond() {
    RespValuePtr response1(new RespValue());
    RespValue* response1_ptr = response1.get();
    EXPECT_CALL(callbacks_, onResponse_(PointeesEq(response1_ptr)));
    pool_callbacks_->onResponse(std::move(response1));
  }

  ConnPool::PoolCallbacks* pool_callbacks_;
  ConnPool::MockPoolRequest pool_request_;
};

TEST_P(RedisSingleServerRequestTest, Success) {
  InSequence s;

  RespValue request;
  makeBulkStringArray(request, {GetParam(), "hello"});
  makeRequest("hello", request);
  EXPECT_NE(nullptr, handle_);

  respond();

  ToLowerTable table;
  std::string lower_command(GetParam());
  table.toLowerCase(lower_command);

  EXPECT_EQ(1UL, store_.counter(fmt::format("redis.foo.command.{}.total", lower_command)).value());
};

TEST_P(RedisSingleServerRequestTest, SuccessMultipleArgs) {
  InSequence s;

  RespValue request;
  makeBulkStringArray(request, {GetParam(), "hello", "123", "world"});
  makeRequest("hello", request);
  EXPECT_NE(nullptr, handle_);

  respond();

  ToLowerTable table;
  std::string lower_command(GetParam());
  table.toLowerCase(lower_command);

  EXPECT_EQ(1UL, store_.counter(fmt::format("redis.foo.command.{}.total", lower_command)).value());
};

TEST_P(RedisSingleServerRequestTest, Fail) {
  InSequence s;

  RespValue request;
  makeBulkStringArray(request, {GetParam(), "hello"});
  makeRequest("hello", request);
  EXPECT_NE(nullptr, handle_);

  fail();
};

TEST_P(RedisSingleServerRequestTest, Cancel) {
  InSequence s;

  RespValue request;
  makeBulkStringArray(request, {GetParam(), "hello"});
  makeRequest("hello", request);
  EXPECT_NE(nullptr, handle_);

  EXPECT_CALL(pool_request_, cancel());
  handle_->cancel();
};

TEST_P(RedisSingleServerRequestTest, NoUpstream) {
  InSequence s;

  RespValue request;
  makeBulkStringArray(request, {GetParam(), "hello"});
  EXPECT_CALL(*conn_pool_, makeRequest("hello", Ref(request), _)).WillOnce(Return(nullptr));
  RespValue response;
  response.type(RespType::Error);
  response.asString() = "no upstream host";
  EXPECT_CALL(callbacks_, onResponse_(PointeesEq(&response)));
  handle_ = splitter_.makeRequest(request, callbacks_);
  EXPECT_EQ(nullptr, handle_);
};

INSTANTIATE_TEST_CASE_P(RedisSingleServerRequestTest, RedisSingleServerRequestTest,
                        testing::ValuesIn(SupportedCommands::simpleCommands()));

INSTANTIATE_TEST_CASE_P(RedisSimpleRequestCommandHandlerMixedCaseTests,
                        RedisSingleServerRequestTest, testing::Values("INCR", "inCrBY"));

TEST_F(RedisSingleServerRequestTest, PingSuccess) {
  InSequence s;

  RespValue request;
  makeBulkStringArray(request, {"ping"});

  RespValue response;
  response.type(RespType::SimpleString);
  response.asString() = "PONG";

  EXPECT_CALL(callbacks_, onResponse_(PointeesEq(&response)));
  handle_ = splitter_.makeRequest(request, callbacks_);
  EXPECT_EQ(nullptr, handle_);
};

TEST_F(RedisSingleServerRequestTest, EvalSuccess) {
  InSequence s;

  RespValue request;
  makeBulkStringArray(request, {"eval", "return {ARGV[1]}", "1", "key", "arg"});
  makeRequest("key", request);
  EXPECT_NE(nullptr, handle_);

  respond();

  ToLowerTable table;
  std::string lower_command("eval");
  table.toLowerCase(lower_command);

  EXPECT_EQ(1UL, store_.counter(fmt::format("redis.foo.command.{}.total", lower_command)).value());
};

TEST_F(RedisSingleServerRequestTest, EvalShaSuccess) {
  InSequence s;

  RespValue request;
  makeBulkStringArray(request, {"EVALSHA", "return {ARGV[1]}", "1", "keykey", "arg"});
  makeRequest("keykey", request);
  EXPECT_NE(nullptr, handle_);

  respond();

  ToLowerTable table;
  std::string lower_command("evalsha");
  table.toLowerCase(lower_command);

  EXPECT_EQ(1UL, store_.counter(fmt::format("redis.foo.command.{}.total", lower_command)).value());
};

TEST_F(RedisSingleServerRequestTest, EvalWrongNumberOfArgs) {
  InSequence s;

  RespValue request;
  RespValue response;
  response.type(RespType::Error);

  response.asString() = "wrong number of arguments for 'eval' command";
  EXPECT_CALL(callbacks_, onResponse_(PointeesEq(&response)));
  makeBulkStringArray(request, {"eval", "return {ARGV[1]}"});
  EXPECT_EQ(nullptr, splitter_.makeRequest(request, callbacks_));

  response.asString() = "wrong number of arguments for 'evalsha' command";
  EXPECT_CALL(callbacks_, onResponse_(PointeesEq(&response)));
  makeBulkStringArray(request, {"evalsha", "return {ARGV[1]}", "1"});
  EXPECT_EQ(nullptr, splitter_.makeRequest(request, callbacks_));
};

TEST_F(RedisSingleServerRequestTest, EvalNoUpstream) {
  InSequence s;

  RespValue request;
  makeBulkStringArray(request, {"eval", "return {ARGV[1]}", "1", "key", "arg"});
  EXPECT_CALL(*conn_pool_, makeRequest("key", Ref(request), _)).WillOnce(Return(nullptr));
  RespValue response;
  response.type(RespType::Error);
  response.asString() = "no upstream host";
  EXPECT_CALL(callbacks_, onResponse_(PointeesEq(&response)));
  handle_ = splitter_.makeRequest(request, callbacks_);
  EXPECT_EQ(nullptr, handle_);
};

class RedisMGETCommandHandlerTest : public RedisCommandSplitterImplTest {
public:
  void setup(uint32_t num_gets, const std::list<uint64_t>& null_handle_indexes) {
    std::vector<std::string> request_strings = {"mget"};
    for (uint32_t i = 0; i < num_gets; i++) {
      request_strings.push_back(std::to_string(i));
    }

    RespValue request;
    makeBulkStringArray(request, request_strings);

    std::vector<RespValue> tmp_expected_requests(num_gets);
    expected_requests_.swap(tmp_expected_requests);
    pool_callbacks_.resize(num_gets);
    std::vector<ConnPool::MockPoolRequest> tmp_pool_requests(num_gets);
    pool_requests_.swap(tmp_pool_requests);
    for (uint32_t i = 0; i < num_gets; i++) {
      makeBulkStringArray(expected_requests_[i], {"get", std::to_string(i)});
      ConnPool::PoolRequest* request_to_use = nullptr;
      if (std::find(null_handle_indexes.begin(), null_handle_indexes.end(), i) ==
          null_handle_indexes.end()) {
        request_to_use = &pool_requests_[i];
      }
      EXPECT_CALL(*conn_pool_, makeRequest(std::to_string(i), Eq(ByRef(expected_requests_[i])), _))
          .WillOnce(DoAll(WithArg<2>(SaveArgAddress(&pool_callbacks_[i])), Return(request_to_use)));
    }

    handle_ = splitter_.makeRequest(request, callbacks_);
  }

  std::vector<RespValue> expected_requests_;
  std::vector<ConnPool::PoolCallbacks*> pool_callbacks_;
  std::vector<ConnPool::MockPoolRequest> pool_requests_;
};

TEST_F(RedisMGETCommandHandlerTest, Normal) {
  InSequence s;

  setup(2, {});
  EXPECT_NE(nullptr, handle_);

  RespValue expected_response;
  expected_response.type(RespType::Array);
  std::vector<RespValue> elements(2);
  elements[0].type(RespType::BulkString);
  elements[0].asString() = "response";
  elements[1].type(RespType::BulkString);
  elements[1].asString() = "5";
  expected_response.asArray().swap(elements);

  RespValuePtr response2(new RespValue());
  response2->type(RespType::BulkString);
  response2->asString() = "5";
  pool_callbacks_[1]->onResponse(std::move(response2));

  RespValuePtr response1(new RespValue());
  response1->type(RespType::BulkString);
  response1->asString() = "response";
  EXPECT_CALL(callbacks_, onResponse_(PointeesEq(&expected_response)));
  pool_callbacks_[0]->onResponse(std::move(response1));

  EXPECT_EQ(1UL, store_.counter("redis.foo.command.mget.total").value());
};

TEST_F(RedisMGETCommandHandlerTest, NormalWithNull) {
  InSequence s;

  setup(2, {});
  EXPECT_NE(nullptr, handle_);

  RespValue expected_response;
  expected_response.type(RespType::Array);
  std::vector<RespValue> elements(2);
  elements[0].type(RespType::BulkString);
  elements[0].asString() = "response";
  expected_response.asArray().swap(elements);

  RespValuePtr response2(new RespValue());
  pool_callbacks_[1]->onResponse(std::move(response2));

  RespValuePtr response1(new RespValue());
  response1->type(RespType::BulkString);
  response1->asString() = "response";
  EXPECT_CALL(callbacks_, onResponse_(PointeesEq(&expected_response)));
  pool_callbacks_[0]->onResponse(std::move(response1));
};

TEST_F(RedisMGETCommandHandlerTest, NoUpstreamHostForAll) {
  // No InSequence to avoid making setup() more complicated.

  RespValue expected_response;
  expected_response.type(RespType::Array);
  std::vector<RespValue> elements(2);
  elements[0].type(RespType::Error);
  elements[0].asString() = "no upstream host";
  elements[1].type(RespType::Error);
  elements[1].asString() = "no upstream host";
  expected_response.asArray().swap(elements);

  EXPECT_CALL(callbacks_, onResponse_(PointeesEq(&expected_response)));
  setup(2, {0, 1});
  EXPECT_EQ(nullptr, handle_);
};

TEST_F(RedisMGETCommandHandlerTest, NoUpstreamHostForOne) {
  InSequence s;

  setup(2, {0});
  EXPECT_NE(nullptr, handle_);

  RespValue expected_response;
  expected_response.type(RespType::Array);
  std::vector<RespValue> elements(2);
  elements[0].type(RespType::Error);
  elements[0].asString() = "no upstream host";
  elements[1].type(RespType::Error);
  elements[1].asString() = "upstream failure";
  expected_response.asArray().swap(elements);

  EXPECT_CALL(callbacks_, onResponse_(PointeesEq(&expected_response)));
  pool_callbacks_[1]->onFailure();
};

TEST_F(RedisMGETCommandHandlerTest, Failure) {
  InSequence s;

  setup(2, {});
  EXPECT_NE(nullptr, handle_);

  RespValue expected_response;
  expected_response.type(RespType::Array);
  std::vector<RespValue> elements(2);
  elements[0].type(RespType::BulkString);
  elements[0].asString() = "response";
  elements[1].type(RespType::Error);
  elements[1].asString() = "upstream failure";
  expected_response.asArray().swap(elements);

  pool_callbacks_[1]->onFailure();

  RespValuePtr response1(new RespValue());
  response1->type(RespType::BulkString);
  response1->asString() = "response";
  EXPECT_CALL(callbacks_, onResponse_(PointeesEq(&expected_response)));
  pool_callbacks_[0]->onResponse(std::move(response1));
};

TEST_F(RedisMGETCommandHandlerTest, InvalidUpstreamResponse) {
  InSequence s;

  setup(2, {});
  EXPECT_NE(nullptr, handle_);

  RespValue expected_response;
  expected_response.type(RespType::Array);
  std::vector<RespValue> elements(2);
  elements[0].type(RespType::Error);
  elements[0].asString() = "upstream protocol error";
  elements[1].type(RespType::Error);
  elements[1].asString() = "upstream failure";
  expected_response.asArray().swap(elements);

  pool_callbacks_[1]->onFailure();

  RespValuePtr response1(new RespValue());
  response1->type(RespType::Integer);
  response1->asInteger() = 5;
  EXPECT_CALL(callbacks_, onResponse_(PointeesEq(&expected_response)));
  pool_callbacks_[0]->onResponse(std::move(response1));
};

TEST_F(RedisMGETCommandHandlerTest, Cancel) {
  InSequence s;

  setup(2, {});
  EXPECT_NE(nullptr, handle_);

  EXPECT_CALL(pool_requests_[0], cancel());
  EXPECT_CALL(pool_requests_[1], cancel());
  handle_->cancel();
};

class RedisMSETCommandHandlerTest : public RedisCommandSplitterImplTest {
public:
  void setup(uint32_t num_sets, const std::list<uint64_t>& null_handle_indexes) {
    std::vector<std::string> request_strings = {"mset"};
    for (uint32_t i = 0; i < num_sets; i++) {
      // key
      request_strings.push_back(std::to_string(i));
      // value
      request_strings.push_back(std::to_string(i));
    }

    RespValue request;
    makeBulkStringArray(request, request_strings);

    std::vector<RespValue> tmp_expected_requests(num_sets);
    expected_requests_.swap(tmp_expected_requests);
    pool_callbacks_.resize(num_sets);
    std::vector<ConnPool::MockPoolRequest> tmp_pool_requests(num_sets);
    pool_requests_.swap(tmp_pool_requests);
    for (uint32_t i = 0; i < num_sets; i++) {
      makeBulkStringArray(expected_requests_[i], {"set", std::to_string(i), std::to_string(i)});
      ConnPool::PoolRequest* request_to_use = nullptr;
      if (std::find(null_handle_indexes.begin(), null_handle_indexes.end(), i) ==
          null_handle_indexes.end()) {
        request_to_use = &pool_requests_[i];
      }
      EXPECT_CALL(*conn_pool_, makeRequest(std::to_string(i), Eq(ByRef(expected_requests_[i])), _))
          .WillOnce(DoAll(WithArg<2>(SaveArgAddress(&pool_callbacks_[i])), Return(request_to_use)));
    }

    handle_ = splitter_.makeRequest(request, callbacks_);
  }

  std::vector<RespValue> expected_requests_;
  std::vector<ConnPool::PoolCallbacks*> pool_callbacks_;
  std::vector<ConnPool::MockPoolRequest> pool_requests_;
};

TEST_F(RedisMSETCommandHandlerTest, Normal) {
  InSequence s;

  setup(2, {});
  EXPECT_NE(nullptr, handle_);

  RespValue expected_response;
  expected_response.type(RespType::SimpleString);
  expected_response.asString() = "OK";

  RespValuePtr response2(new RespValue());
  response2->type(RespType::SimpleString);
  response2->asString() = "OK";
  pool_callbacks_[1]->onResponse(std::move(response2));

  RespValuePtr response1(new RespValue());
  response1->type(RespType::SimpleString);
  response1->asString() = "OK";
  EXPECT_CALL(callbacks_, onResponse_(PointeesEq(&expected_response)));
  pool_callbacks_[0]->onResponse(std::move(response1));

  EXPECT_EQ(1UL, store_.counter("redis.foo.command.mset.total").value());
};

TEST_F(RedisMSETCommandHandlerTest, NoUpstreamHostForAll) {
  // No InSequence to avoid making setup() more complicated.

  RespValue expected_response;
  expected_response.type(RespType::Error);
  expected_response.asString() = "finished with 2 error(s)";

  EXPECT_CALL(callbacks_, onResponse_(PointeesEq(&expected_response)));
  setup(2, {0, 1});
  EXPECT_EQ(nullptr, handle_);
};

TEST_F(RedisMSETCommandHandlerTest, NoUpstreamHostForOne) {
  InSequence s;

  setup(2, {0});
  EXPECT_NE(nullptr, handle_);

  RespValue expected_response;
  expected_response.type(RespType::Error);
  expected_response.asString() = "finished with 1 error(s)";

  RespValuePtr response2(new RespValue());
  response2->type(RespType::SimpleString);
  response2->asString() = "OK";
  EXPECT_CALL(callbacks_, onResponse_(PointeesEq(&expected_response)));
  pool_callbacks_[1]->onResponse(std::move(response2));
};

TEST_F(RedisMSETCommandHandlerTest, Cancel) {
  InSequence s;

  setup(2, {});
  EXPECT_NE(nullptr, handle_);

  EXPECT_CALL(pool_requests_[0], cancel());
  EXPECT_CALL(pool_requests_[1], cancel());
  handle_->cancel();
};

TEST_F(RedisMSETCommandHandlerTest, WrongNumberOfArgs) {
  InSequence s;

  RespValue response;
  response.type(RespType::Error);
  response.asString() = "wrong number of arguments for 'mset' command";
  EXPECT_CALL(callbacks_, onResponse_(PointeesEq(&response)));
  RespValue request;
  makeBulkStringArray(request, {"mset", "foo", "bar", "fizz"});
  EXPECT_EQ(nullptr, splitter_.makeRequest(request, callbacks_));
};

class RedisSplitKeysSumResultHandlerTest : public RedisCommandSplitterImplTest,
                                           public testing::WithParamInterface<std::string> {
public:
  void setup(uint32_t num_commands, const std::list<uint64_t>& null_handle_indexes) {
    std::vector<std::string> request_strings = {GetParam()};
    for (uint32_t i = 0; i < num_commands; i++) {
      request_strings.push_back(std::to_string(i));
    }

    RespValue request;
    makeBulkStringArray(request, request_strings);

    std::vector<RespValue> tmp_expected_requests(num_commands);
    expected_requests_.swap(tmp_expected_requests);
    pool_callbacks_.resize(num_commands);
    std::vector<ConnPool::MockPoolRequest> tmp_pool_requests(num_commands);
    pool_requests_.swap(tmp_pool_requests);
    for (uint32_t i = 0; i < num_commands; i++) {
      makeBulkStringArray(expected_requests_[i], {GetParam(), std::to_string(i)});
      ConnPool::PoolRequest* request_to_use = nullptr;
      if (std::find(null_handle_indexes.begin(), null_handle_indexes.end(), i) ==
          null_handle_indexes.end()) {
        request_to_use = &pool_requests_[i];
      }
      EXPECT_CALL(*conn_pool_, makeRequest(std::to_string(i), Eq(ByRef(expected_requests_[i])), _))
          .WillOnce(DoAll(WithArg<2>(SaveArgAddress(&pool_callbacks_[i])), Return(request_to_use)));
    }

    handle_ = splitter_.makeRequest(request, callbacks_);
  }

  std::vector<RespValue> expected_requests_;
  std::vector<ConnPool::PoolCallbacks*> pool_callbacks_;
  std::vector<ConnPool::MockPoolRequest> pool_requests_;
};

TEST_P(RedisSplitKeysSumResultHandlerTest, Normal) {
  InSequence s;

  setup(2, {});
  EXPECT_NE(nullptr, handle_);

  RespValue expected_response;
  expected_response.type(RespType::Integer);
  expected_response.asInteger() = 2;

  RespValuePtr response2(new RespValue());
  response2->type(RespType::Integer);
  response2->asInteger() = 1;
  pool_callbacks_[1]->onResponse(std::move(response2));

  RespValuePtr response1(new RespValue());
  response1->type(RespType::Integer);
  response1->asInteger() = 1;
  EXPECT_CALL(callbacks_, onResponse_(PointeesEq(&expected_response)));
  pool_callbacks_[0]->onResponse(std::move(response1));

  EXPECT_EQ(1UL, store_.counter("redis.foo.command." + GetParam() + ".total").value());
};

TEST_P(RedisSplitKeysSumResultHandlerTest, NormalOneZero) {
  InSequence s;

  setup(2, {});
  EXPECT_NE(nullptr, handle_);

  RespValue expected_response;
  expected_response.type(RespType::Integer);
  expected_response.asInteger() = 1;

  RespValuePtr response2(new RespValue());
  response2->type(RespType::Integer);
  response2->asInteger() = 0;
  pool_callbacks_[1]->onResponse(std::move(response2));

  RespValuePtr response1(new RespValue());
  response1->type(RespType::Integer);
  response1->asInteger() = 1;
  EXPECT_CALL(callbacks_, onResponse_(PointeesEq(&expected_response)));
  pool_callbacks_[0]->onResponse(std::move(response1));

  EXPECT_EQ(1UL, store_.counter("redis.foo.command." + GetParam() + ".total").value());
};

TEST_P(RedisSplitKeysSumResultHandlerTest, NoUpstreamHostForAll) {
  // No InSequence to avoid making setup() more complicated.

  RespValue expected_response;
  expected_response.type(RespType::Error);
  expected_response.asString() = "finished with 2 error(s)";

  EXPECT_CALL(callbacks_, onResponse_(PointeesEq(&expected_response)));
  setup(2, {0, 1});
  EXPECT_EQ(nullptr, handle_);
};

INSTANTIATE_TEST_CASE_P(RedisSplitKeysSumResultHandlerTest, RedisSplitKeysSumResultHandlerTest,
                        testing::ValuesIn(SupportedCommands::hashMultipleSumResultCommands()));

} // namespace CommandSplitter
} // namespace RedisProxy
} // namespace NetworkFilters
} // namespace Extensions
} // namespace Envoy
